module Sinbotra::Messenger
  class Bot
    GET_STARTED_ID = "GET_STARTED"

    class << self
      attr_reader :_listeners, :_default_listener, :_conversations, :_dont_understand_phrases, :get_started_method

      Listener     = Struct.new(:matcher, :method_name)
      Conversation = Struct.new(:id, :klass)

      def get_started(method_name)
        @get_started_method = method_name
        Sinbotra::Messenger::Platform.get_started(GET_STARTED_ID)
      end

      def listeners(methods)
        @_listeners = methods.map { |name, matcher| Listener.new(matcher, name) }
      end

      def default_listener(listener)
        @_default_listener = listener
      end

      def conversations(conversation_map)
        c = conversation_map.each_with_object({}) { |(id, klass), h|
          h[id] = Conversation.new(id, klass)
        }
        @_conversations = c
      end

      def dont_understand_phrases(phrases)
        @_dont_understand_phrases = phrases
      end
    end

    attr_reader :current_user
    attr_reader :message
    attr_reader :platform
    attr_reader :get_started_method

    def listeners;          self.class._listeners;         end
    def default_listener;   self.class._default_listener;  end
    def conversations;      self.class._conversations;     end
    def get_started_method; self.class.get_started_method; end

    include Sinbotra::Messenger

    def initialize(message)
      @current_user = UserPresenter.new(message.sender)
      @message      = MessagePresenter.new(message)
      @platform     = Platform.new(@current_user.id)
      @phrases      = setup_phrases
      log_message!
    end

    def respond
      $logger.debug("Responding...")
      in_conversation? ? send_to_conversation! : send_to_listeners!
    end

    def in_conversation?; current_user.in_conversation?; end

    def start_conversation(convo_id)
      current_user.start_conversation(convo_id)
      convo = make_conversation(convo_id)
      $logger.debug("Start Conversation: #{convo}")
      convo.start
    end

    def make_conversation(convo_id)
      convo = conversations[convo_id]
      convo.klass.new(self, platform)
    end

    def send_to_conversation!
      $logger.debug("Send message to conversation")
      id    = current_user.conversation.id
      convo = make_conversation(id)
      convo.continue_dialogue
    end

    def send_to_listeners!
      $logger.debug("Send message through handler stack")
      listener = listeners.find { |l| matches_message?(l) }
      if listener.nil?
        get_started_message? ? handle_get_started! : handle_default!
      else
        send(listener.method_name, message)
      end
    end

    def get_started_message?
      pb = message.postback
      pb.to_s == self.class::GET_STARTED_ID
    end

    def handle_get_started!
      send(get_started_method, message)
    end

    def handle_default!
      send(default_listener, message)
    end

    def has_conversation?(intent)
      # TODO: Figure this out ;)
      false
    end

    def get_intent(msg)
      # TODO: Send msg to NLP API
    end

    def next_dont_understand_phrase
      return "I don't understand" unless @phrases.any?
      @phrase_index ||= 0
      phrase = @phrases[@phrase_index % @phrases.size]
      @phrase_index += 1
      phrase
    end

    private

    def matches_message?(listener)
      if txt = message.text
        return true if txt.match(listener.matcher)
      elsif pb = message.postback
        return true if pb.to_s.match(listener.matcher)
      end
      return false
    end

    def setup_phrases
      phrases = self.class._dont_understand_phrases || []
      phrases.shuffle
    end

    def log_message!
      Sinbotra::MessageStore.log_in_message!(:facebook, message, message.sender)
    end
  end
end
